//-----------------------------------------------------------------------------
// vtSerialComm.cpp: implementation of the vtSerialComm class.
//
// Copyright (c) by Volker Typke 2001 (www.typke.de).
//
// Please read the comments on how to use the functions and also consult the
// help files provided with Visual C++.
//
// Revision history:
//
// Revision		Date		Changes
// ----------------------------------------------------------------------------
// 1.0.1	Jan 06. 2002	AbortRead(), AbortWrite() and FlushWriteBuffer()
//							added.
// 1.0		Dec 20, 2001	Initial version.
//-----------------------------------------------------------------------------

#include "stdafx.h"
#include "vtSerialComm.h"


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

vtSerialComm::vtSerialComm()
{
	m_cXonChar = 17;
	m_cXoffChar = 19;
	m_nXonLim = 1024;
	m_nXoffLim = 1024;
}

vtSerialComm::~vtSerialComm()
{
	Close ();
}

//-----------------------------------------------------------------------------
// OpenCommPort()
//
// Opens the specified serial communications port for generic read/write access
// Valid port names are "COM1", "COM2", and so on.
//
// Parameters:
// 
// Returns:
// - commport_ERROR_SUCCESS	if the specified serial communications port could
//							be opened successfully
// - commport_ERROR_OPEN	if the specified serial communications port could
//							not be opened. Further information about the reason
//							why the operation failed can be obtained by calling
//							GetLastSystemError().
//-----------------------------------------------------------------------------
int vtSerialComm::Open (LPCTSTR lpszCommPort, DWORD dwBaudRate, BYTE byByteSize, BYTE byStopBits, BYTE byParity, BYTE byFlowControl)
{
	CString string;					// generic string for error messages, and so on

	// reset member variables
	m_nBytesRead = 0;
	m_nBytesWritten = 0;
	m_nSystemError = 0;

	// open serial communications port
	m_hCommPort = CreateFile (
		lpszCommPort,					// file name: some serial communictaions port
		GENERIC_READ | GENERIC_WRITE,	// access mode: read/write
		0,								// share mode: exclusive access
		NULL,							// security atrributes: none
		OPEN_EXISTING,					// comm devices must use OPEN_EXISTING
		FILE_FLAG_OVERLAPPED,			// open for overlapped I/O
		NULL);							// use no file template

	if (m_hCommPort == INVALID_HANDLE_VALUE)
	{
		m_nSystemError = GetLastError ();
		return commport_ERROR_OPEN;
	}

	// initialize timeouts (no timeouts are used)
	COMMTIMEOUTS cto = {0, 0, 0, 0, 0};

	if (!SetCommTimeouts (m_hCommPort, &cto))
	{
		// error handling
		return commport_ERROR_OPEN;
	}

	// initialize serial communications port
	// (in order not to have to fill in the whole DCB structure, let's get the
	//  current configuration and change only things that really need to be changed)
	if (!GetCommState (m_hCommPort, &m_dcb))
	{
		return commport_ERROR_GET_DCB;
	}

	// initialize device control structure
	switch (byFlowControl)
	{
	case commport_FLOW_NONE:							// no flow control
	case commport_FLOW_SOFT:							// flow control by software
		m_dcb.fRtsControl		= RTS_CONTROL_DISABLE;	// disable RTS line
		m_dcb.fDtrControl		= DTR_CONTROL_DISABLE;	// disable DTR line
		m_dcb.fOutxCtsFlow		= 0;					// ignore CTS for output flow control
		m_dcb.fOutxDsrFlow		= 0;					// ignore DSR for output flow control
		m_dcb.fDsrSensitivity	= 0;					// ignore DSR for input flow control
		m_dcb.fOutX				= 0;					// output flow control: turn off Xon/Xoff
		m_dcb.fInX				= 0;					// input flow control: turn off Xon/Xoff
		break;
	case commport_FLOW_HARD:
		m_dcb.fRtsControl		= RTS_CONTROL_HANDSHAKE;// use RTS/CTS handshaking
		m_dcb.fDtrControl		= DTR_CONTROL_ENABLE;	// always enable DTR line
		m_dcb.fOutxCtsFlow		= 0;					// ignore CTS for output flow control
		m_dcb.fOutxDsrFlow		= 0;					// ignore DSR for output flow control
		m_dcb.fDsrSensitivity	= 0;					// ignore DSR for input flow control
		m_dcb.fOutX				= 1;					// output flow control: turn off Xon/Xoff
		m_dcb.fInX				= 1;					// input flow control: turn off Xon/Xoff
		m_dcb.XonChar			= m_cXonChar;			// specify Xon character
		m_dcb.XoffChar			= m_cXoffChar;			// specify Xoff character
		m_dcb.XonLim			= m_nXonLim;			// specify Xon limit
		m_dcb.XoffLim			= m_nXoffLim;			// sepcify Xoff limit
		break;
	case commport_FLOW_XONXOFF:
		m_dcb.fRtsControl		= RTS_CONTROL_DISABLE;	// disable RTS line
		m_dcb.fDtrControl		= DTR_CONTROL_DISABLE;	// disable DTR line
		m_dcb.fOutxCtsFlow		= 1;					// ignore CTS for output flow control
		m_dcb.fOutxDsrFlow		= 0;					// ignore DSR for output flow control
		m_dcb.fDsrSensitivity	= 0;					// ignore DSR for input flow control
		m_dcb.fOutX				= 0;					// output flow control: turn off Xon/Xoff
		m_dcb.fInX				= 0;					// input flow control: turn off Xon/Xoff
		break;
	default:
		return commport_ERROR_INVALID_FLOW;
	}
	m_dcb.fParity			= 1;				// enable parity checking
	m_dcb.fNull				= 0;				// do not discard NULL bytes when received
	m_dcb.fAbortOnError		= 1;				// abort on error
	m_dcb.BaudRate			= dwBaudRate;		// baud rate
	m_dcb.ByteSize			= byByteSize;		// byte size
	m_dcb.StopBits			= byStopBits;		// stop bits
	m_dcb.Parity			= byParity;			// parity

	// now set the new attributes
	if (!SetCommState (m_hCommPort, &m_dcb))
	{
		return commport_ERROR_SET_DCB;
	}


	// create event for overlapped I/O (read)
	m_hReadEvent = CreateEvent (	NULL,		// no security attributes
									TRUE,		// manual reset
									TRUE,		// initially signaled
									NULL);		// no name

	if (m_hReadEvent == INVALID_HANDLE_VALUE)
	{
		return commport_ERROR_CREATE_EVENT;
	}


	// create event for overlapped I/O (write)
	m_hWriteEvent = CreateEvent (	NULL,		// no security attributes
									TRUE,		// manual reset
									TRUE,		// initially signaled
									NULL);		// no name

	if (m_hWriteEvent == INVALID_HANDLE_VALUE)
	{
		return commport_ERROR_CREATE_EVENT;
	}

	return commport_ERROR_SUCCESS;
}


//-----------------------------------------------------------------------------
// GetLastSystemError()
//
// Returns the value of the last sytsem error code that was stored by a
// vtSerialComm function. This provides more detailed information on why a
// certain vtSerialComm function failed.
// For example, OpenCommPort() will return commport_ERROR_OPEN, if it fails.
// Calling GetLastSystemError right after OpenCommPort will return the system
// error that OpenCommPort encountered, e. g. ERROR_ACCESS_DENIED if the
// specified serial communications port was already in use by another
// application or ERROR_FILE_NOT_FOUND if the specified serial communications
// port simply does not exist.
//-----------------------------------------------------------------------------
DWORD vtSerialComm::GetLastSystemError()
{
	return m_nSystemError;
}

//-----------------------------------------------------------------------------
// ReadOverlapped()
//
// This function starts a read operation to read the given number of bytes into
// the given buffer. The function returns immediately, that is it does not
// wait for the read operation to be finished first. Instead, the read
// operation is performed in the background giving the calling function time to
// perform other things in parallel while waiting for the overlapped read
// operation to be completed. You can check if the read operation is
// complete with calls to BytesRead(), which will return the number of bytes
// read since the last call to ReadOverlapped() or Read(). Once the number is
// equal to the number of bytes you specified to be read during the call to
// ReadOverlapped() or Read(), you know that the read operation is finished.
//
// Do not call ReadOverlapped() until any other read operation is complete.
//
// Returns:
// - commport_ERROR_SUCCESS if no error occurs (the system error
//							ERROR_IO_PENDING is expected and treated as no
//							error)
// - commport_ERROR_READ	if an error other than ERROR_IO_PENDING occurs. You
//							can call GetLastSystemError() to find out why
//							the function failed.
//-----------------------------------------------------------------------------
int vtSerialComm::ReadOverlapped(LPVOID lpBuffer, DWORD nLength)
{
	m_ovRead.Offset = 0;
	m_ovRead.OffsetHigh = 0;
	m_ovRead.hEvent = m_hReadEvent;

	m_nBytesRead = 0;

	if (!ReadFile (m_hCommPort, lpBuffer, nLength, &m_nBytesRead, &m_ovRead))
	{
		m_nSystemError = GetLastError ();

		if (m_nSystemError != ERROR_IO_PENDING)
		{
			return commport_ERROR_READ;
		}
	}

	return commport_ERROR_SUCCESS;
}

//-----------------------------------------------------------------------------
// BytesRead()
//
// This function returns the number of bytes read since the last call to
// ReadOverlapped() or Read(). You can use this function to check if an
// overlapped read operation is complete. This is the case when BytesRead()
// returns the number of bytes as specified during the call to
// ReadOverlapped() or Read().
//-----------------------------------------------------------------------------
DWORD vtSerialComm::BytesRead()
{
	GetOverlappedResult (m_hCommPort, &m_ovRead, &m_nBytesRead, FALSE);
	return m_nBytesRead;
}

//-----------------------------------------------------------------------------
// Read()
//
// This function reads the specified number of bytes into the given buffer.
// It returns as soon as the specified number of bytes was received or when an
// error occurrs.
//
// Do not call ReadOverlapped() until any other read operation is complete.
//
// Returns:
// - commport_ERROR_SUCCESS if no error occurs,
// - commport_ERROR_READ	if an error occurs. You can call
//							GetLastSystemError() to find out why
//							the function failed.
//-----------------------------------------------------------------------------
int vtSerialComm::Read(LPVOID lpBuffer, DWORD nLength)
{
	m_ovRead.Offset		= 0;
	m_ovRead.OffsetHigh	= 0;
	m_ovRead.hEvent		= m_hReadEvent;

	m_nBytesRead = 0;

	if (!ReadFile (m_hCommPort, lpBuffer, nLength, &m_nBytesRead, &m_ovRead))
	{
		if (!GetOverlappedResult (m_hCommPort, &m_ovRead, &m_nBytesRead, TRUE))
		{
			// error handling
			m_nSystemError = GetLastError ();
			return commport_ERROR_READ;
		}
	}

	return commport_ERROR_SUCCESS;
}

//-----------------------------------------------------------------------------
// Close()
//
// This function closes all handles used by this object.
//-----------------------------------------------------------------------------
void vtSerialComm::Close()
{
	CloseHandle (m_hReadEvent);
	CloseHandle (m_hWriteEvent);
	CloseHandle (m_hCommPort);
}

//-----------------------------------------------------------------------------
// WriteOverlapped()
//
// This function starts a write operation to write the given number of bytes
// from the given buffer to the serial communications port.
// The function returns immediately, that is it does not wait for the write
// operation to be finished first. Instead, the write operation is performed in
// the background giving the calling function time to perform other things in
// parallel while waiting for the overlapped write operation to be completed.
// You can check if the write operation is complete with calls to
// BytesWritten(), which will return the number of bytes written to the serial
// communications port since the last call to WriteOverlapped() or Write().
// Once the number is equal to the number of bytes you specified to be written
// when you called WriteOverlapped() or Write(), the write operation is
// finished.
//
// Returns:
// - commport_ERROR_SUCCESS if no error occurs (the system error
//							ERROR_IO_PENDING is expected and treated as no
//							error)
// - commport_ERROR_WRITE	if an error other than ERROR_IO_PENDING occurs. You
//							can call GetLastSystemError() to find out why
//							the function failed.
//-----------------------------------------------------------------------------
int vtSerialComm::WriteOverlapped(LPVOID lpBuffer, DWORD nLength)
{
	m_ovWrite.Offset		= 0;
	m_ovWrite.OffsetHigh	= 0;
	m_ovWrite.hEvent		= m_hWriteEvent;

	m_nBytesWritten = 0;

	if (!WriteFile (m_hCommPort, lpBuffer, nLength, &m_nBytesWritten, &m_ovWrite))
	{
		m_nSystemError = GetLastError ();

		if (m_nSystemError != ERROR_IO_PENDING)
		{
			return commport_ERROR_WRITE;
		}
	}

	return commport_ERROR_SUCCESS;
}

//-----------------------------------------------------------------------------
// Write()
//
// This function writes the sepcified number of bytes from the given buffer to
// the serial communications port. It returns after the operation is complete.
//
// Returns:
// - commport_ERROR_SUCCESS if no error occurs
// - commport_ERROR_WRITE	if an error occurs. You can call
//							GetLastSystemError() to find out why
//							the function failed.
//-----------------------------------------------------------------------------
int vtSerialComm::Write(LPVOID lpBuffer, DWORD nLength)
{
	m_ovWrite.Offset		= 0;
	m_ovWrite.OffsetHigh	= 0;
	m_ovWrite.hEvent		= m_hWriteEvent;

	m_nBytesWritten = 0;

	if (!WriteFile (m_hCommPort, lpBuffer, nLength, &m_nBytesWritten, &m_ovWrite))
	{
		if (!GetOverlappedResult (m_hCommPort, &m_ovWrite, &m_nBytesWritten, TRUE))
		{
			// error handling
			m_nSystemError = GetLastError ();
			return commport_ERROR_WRITE;
		}
	}

	return commport_ERROR_SUCCESS;
}

//-----------------------------------------------------------------------------
// BytesWritten()
//
// This function returns the number of bytes written to the serial
// communications port since the last call to WriteOverlapped() or Write().
// you can use this function to determin whether an overlapped write operation
// is complete or not. If the operation is complete, the number returned by
// this function will be equal to the number of bytes to send when you called
// WriteOverlapped() or Write().
//-----------------------------------------------------------------------------
DWORD vtSerialComm::BytesWritten()
{
	GetOverlappedResult (m_hCommPort, &m_ovWrite, &m_nBytesWritten, FALSE);
	return m_nBytesWritten;
}


//-----------------------------------------------------------------------------
// SetXonLim()
//
// Sets the Xon limit. This needs to be done before opening the serial
// communications port, but only if you will use XON/XOFF flow control.
//-----------------------------------------------------------------------------
void vtSerialComm::SetXonLim(WORD nXonLim)
{
	m_nXonLim = nXonLim;
}

//-----------------------------------------------------------------------------
// SetXoffLim()
//
// Sets the Xoff limit. This needs to be done before opening the serial
// communications port, but only if you will use XON/XOFF flow control.
//-----------------------------------------------------------------------------
void vtSerialComm::SetXoffLim(WORD nXoffLim)
{
	m_nXoffLim = nXoffLim;
}

//-----------------------------------------------------------------------------
// SetXonChar()
//
// Sets the Xon character. This needs to be done before opening the serial
// communications port, but only if you will use XON/XOFF flow control.
//-----------------------------------------------------------------------------
void vtSerialComm::SetXonChar(char cXonChar)
{
	m_cXonChar = cXonChar;
}

//-----------------------------------------------------------------------------
// SetXoffChar()
//
// Sets the Xoff character. This needs to be done before opening the serial
// communications port, but only if you will use XON/XOFF flow control.
//-----------------------------------------------------------------------------
void vtSerialComm::SetXoffChar(char cXoffChar)
{
	m_cXoffChar = cXoffChar;
}

//-----------------------------------------------------------------------------
// AbortRead()
//
// This function can be used to terminate all outstanding overlapped read
// operations and to clear the serial communications driver's input buffer.
// The function returns immediately, even if the read operation has not been
// completed.
//
// Parameters:
// - dwFlags		Specifies the action to take. It can be either one of the
//					values given below or the OR-combination of both.
//					The default parameter is PURGE_RXABORT to terminate the
//					read operation without clearing the input buffer.
//
//	PURGE_RXABORT	terminates all outstanding read operations.
//	PURGE_RXCLEAR	clears the input buffer of the device driver.
//
// Returns:
// - comport_ERROR_SUCCESS, if the function succeeds,
// - commport_ERROR_ABORT_READ, if the function failed.
//
// To get more information on why the function failed, call
// GetLastSystemError().
//-----------------------------------------------------------------------------
int vtSerialComm::AbortRead(DWORD dwFlags)
{
	if (!PurgeComm (m_hCommPort, dwFlags))
	{
		// error handling
		m_nSystemError = GetLastError ();
		return commport_ERROR_ABORT_READ;
	}

	return commport_ERROR_SUCCESS;
}

//-----------------------------------------------------------------------------
// AbortWrite()
//
// This function can be used to terminate all outstanding overlapped write
// operations and to clear the serial communications driver's output buffer.
// The function returns immediately, even if the write operation has not been
// completed. Any characters that have not been sent when calling this function
// will not be transmitted. To empty the output buffer while ensuring that the
// contents are transmitted, call FlushWriteBuffer().
//
// Parameters:
// - dwFlags		Specifies the action to take. It can be either one of the
//					values given below or the OR-combination of both.
//					The default parameter is PURGE_TXABORT to terminate the
//					write operation without clearing the output buffer.
//
//	PURGE_TXABORT	terminates all outstanding write operations.
//	PURGE_TXCLEAR	clears the output buffer of the device driver.
//
// Returns:
// - comport_ERROR_SUCCESS, if the function succeeds,
// - commport_ERROR_ABORT_WRITE, if the function failed.
//
// To get more information on why the function failed, call
// GetLastSystemError().
//-----------------------------------------------------------------------------
int vtSerialComm::AbortWrite(DWORD dwFlags)
{
	if (!PurgeComm (m_hCommPort, dwFlags))
	{
		// error handling
		m_nSystemError = GetLastError ();
		return commport_ERROR_ABORT_WRITE;
	}

	return commport_ERROR_SUCCESS;
}

//-----------------------------------------------------------------------------
// FlushWriteBuffer()
//
// This function can be used to empty the serial communications driver's
// output buffer. This is a synchroneous operation that is subject to flow
// control but not write time-outs, and it will not return until ALL pending
// write operations have been transmitted.
//
// Returns:
// - comport_ERROR_SUCCESS, if the function succeeds,
// - commport_ERROR_FLUSH_WRITE_BUFFER, if the function failed.
//
// To get more information on why the function failed, call
// GetLastSystemError().
//-----------------------------------------------------------------------------
int vtSerialComm::FlushWriteBuffer()
{
	if (!FlushFileBuffers (m_hCommPort))
	{
		// error handling
		m_nSystemError = GetLastError ();
		return commport_ERROR_FLUSH_WRITE_BUFFER;
	}

	return commport_ERROR_SUCCESS;
}
